{- $Id: PioneerBridge.hs,v 1.1.1.1 2002/06/26 21:40:30 henrik Exp $
******************************************************************************
*                          A F R O B / P I O N E E R                         *
*                                                                            *
*	Module:		PioneerBridge					     *
*	Purpose:	The Haskell part of the PioneerBridge.		     *
*	Author:		Henrik Nilsson					     *
*									     *
******************************************************************************
-}

module PioneerBridge (
    pioneerDiameter,
    pioneerMaxAcceleration,
    pioneerMaxSpeed,
    pioneerServerRate,
    pioneerResetOdometry,
    pioneerGetPos,
    pioneerSetVelTR,
    pioneerSetVelTRGetPos,
    pioneerSetVelDiff,
    pioneerSetVelDiffGetPos,
    pioneerStop,
    pioneerStopGetPos,
    startServers
) where

import Monad (when)
import Foreign
import AFrobPhysicalDimensions


------------------------------------------------------------------------------
-- Primitives (from PioneerBridgePrim)
------------------------------------------------------------------------------

foreign import "pioneerDiameter_prim" pioneerDiameter :: Double
foreign import "pioneerMaxAcceleration_prim" pioneerMaxAcceleration :: Double
foreign import "pioneerMaxSpeed_prim" pioneerMaxSpeed :: Double
foreign import pioneerServerRate_prim :: IO Double
foreign import pioneerResetOdometry_prim :: IO Bool
foreign import pioneerGetPos_prim ::
    Ptr Double -> Ptr Double -> Ptr Double -> IO Bool
foreign import pioneerSetVelTR_prim :: Double -> Double -> IO Bool
foreign import pioneerSetVelTRGetPos_prim ::
    Double -> Double -> Ptr Double -> Ptr Double -> Ptr Double -> IO Bool
foreign import pioneerSetVelDiff_prim :: Double -> Double -> IO Bool
foreign import pioneerSetVelDiffGetPos_prim ::
    Double -> Double -> Ptr Double -> Ptr Double -> Ptr Double -> IO Bool
foreign import pioneerStop_prim :: IO Bool
foreign import pioneerStopGetPos_prim  :: 
    Ptr Double -> Ptr Double -> Ptr Double -> IO Bool
foreign import startServers_prim :: IO Bool


------------------------------------------------------------------------------
-- Exposed interface
------------------------------------------------------------------------------

-- All these raise an exception on error.

-- Gets the current server rate.

pioneerServerRate :: IO Frequency
pioneerServerRate = pioneerServerRate_prim


-- Reset odometry to positiom (0,0) and heading 0.

pioneerResetOdometry :: IO ()
pioneerResetOdometry = do
    success <- pioneerResetOdometry_prim
    when (not success) (ioError (pbExc "pioneerResetOdometry" "failed"))
    return ()


-- Get current position according to odometry.
--
-- Returns a three tuple:
-- #1 .........	Current X position [m].
-- #2 .........	Current y position [m].
-- #3 .........	Current heading [rad]

pioneerGetPos :: IO (Position, Position, Heading)
pioneerGetPos = alloca3 $ \xp yp thetap -> do
    success <- pioneerGetPos_prim xp yp thetap
    when (not success) (ioError (pbExc "pioneerGetPos" "failed"))
    x     <- peek xp
    y     <- peek yp
    theta <- peek thetap
    return (x, y, theta)


-- Set translational and rotational velocity.
--
-- v ..........	Translational velocity [m/s].
-- rv .........	Rotational velocity [rad/s].

pioneerSetVelTR :: Velocity -> RotVel -> IO ()
pioneerSetVelTR v rv = do
    success <- pioneerSetVelTR_prim v rv
    when (not success) (ioError (pbExc "pioneerSetVelTr" "failed"))
    return ()


-- Set translational and rotational velocity and get current position
-- according to odometry.
--
-- v ..........	Translational velocity [m/s].
-- rv .........	Rotational velocity [rad/s].
--
-- Returns a three tuple:
-- #1 .........	Current X position [m].
-- #2 .........	Current y position [m].
-- #3 .........	Current heading [rad]

pioneerSetVelTRGetPos :: Velocity -> RotVel -> IO (Position, Position, Heading)
pioneerSetVelTRGetPos v rv = alloca3 $ \xp yp thetap -> do
    success <- pioneerSetVelTRGetPos_prim v rv xp yp thetap
    when (not success) (ioError (pbExc "pioneerSetVelTrGetPos" "failed"))
    x     <- peek xp
    y     <- peek yp
    theta <- peek thetap
    return (x, y, theta)
	


-- Set left and right wheel peripheral velocity.
--
-- lwv ........	Left wheel peripheral velocity [m/s].
-- rwv ........	Right wheel peripheral velocity [m/s].

pioneerSetVelDiff :: Velocity -> Velocity -> IO ()
pioneerSetVelDiff lwv rwv = do
    success <- pioneerSetVelDiff_prim lwv rwv
    when (not success) (ioError (pbExc "pioneerSetVelDiff" "failed"))
    return ()


-- Sets left and right wheel peripheral velocity and get current position
-- according to odometry.
--
-- lwv ........	Left wheel peripheral velocity [m/s].
-- rwv ........	Right wheel peripheral velocity [m/s].
--
-- Returns a three tuple:
-- #1 .........	Current X position [m].
-- #2 .........	Current y position [m].
-- #3 .........	Current heading [rad]

pioneerSetVelDiffGetPos ::
    Velocity -> Velocity -> IO (Position, Position, Heading)
pioneerSetVelDiffGetPos lwv rwv = alloca3 $ \xp yp thetap -> do
    success <- pioneerSetVelDiffGetPos_prim lwv rwv xp yp thetap
    when (not success) (ioError (pbExc "pioneerSetVelDiffGetPos" "failed"))
    x     <- peek xp
    y     <- peek yp
    theta <- peek thetap
    return (x, y, theta)


-- Stop the pioneer.

pioneerStop :: IO ()
pioneerStop = do
    success <- pioneerStop_prim
    when (not success) (ioError (pbExc "pioneerStop" "failed"))
    return ()


-- Stop the pioneer and get current position.
--
-- Returns a three tuple:
-- #1 .........	Current X position [m].
-- #2 .........	Current y position [m].
-- #3 .........	Current heading [rad]

pioneerStopGetPos :: IO (Position, Position, Heading)
pioneerStopGetPos = alloca3 $ \xp yp thetap -> do
    success <- pioneerStopGetPos_prim xp yp thetap
    when (not success) (ioError (pbExc "pioneerStopGetPos" "failed"))
    x     <- peek xp
    y     <- peek yp
    theta <- peek thetap
    return (x, y, theta)


-- Initialize clients and start servers.

startServers :: IO ()
startServers = do
    success <- startServers_prim
    when (not success) (ioError (pbExc "startServers" "failed"))
    return ()


------------------------------------------------------------------------------
-- Utilities
------------------------------------------------------------------------------

alloca2 :: (Storable a, Storable b) => (Ptr a -> Ptr b -> IO c) -> IO c
alloca2 f = alloca (\ptr -> alloca (f ptr))

alloca3 :: (Storable a, Storable b, Storable c)
    => (Ptr a -> Ptr b -> Ptr c -> IO d) -> IO d
alloca3 f = alloca (\ptr -> alloca2 (f ptr))

pbExc fun msg = userError ("PioneerDriver." ++ fun ++ ": " ++ msg)
