{- $Id: PioneerDriver.hs,v 1.1.1.1 2002/06/26 21:40:30 henrik Exp $
******************************************************************************
*                          A F R O B / P I O N E E R                         *
*                                                                            *
*	Module:		PioneerDriver					     *
*	Purpose:	Quick and dirty reactimate hask for Pioneers	     *
*	Author:		Henrik Nilsson					     *
*									     *
******************************************************************************
-}

module PioneerDriver (
    reactimatePioneer,
    reactimatePioneerVS
) where

import Posix (SysVar(..), ProcessTimes, ClockTick,
              getSysVar, getProcessTimes, elapsedTime)
import IOExts (newIORef, readIORef, writeIORef)
import AFRP (ST, PrimFilter, SysSampler, SysRenderer, reactimate)
import AFRPTask (repeatUntil, forAll)
import AFVision
import AFrobPhysicalDimensions
import AFrobRobotIO
import PioneerBridge
import PioneerIO

-- ToDo:
-- * Text console output to stderr? (sysRenderer) Maybe not.
-- * Diagnostic output a la Cye? (Maybe diagnosticHandle) as first arg to
--   reactimate pioneer.
-- * For debugging purposes, we might like to keep track of the absolute
--   time (since system start). Again, see Cye.
-- * The reactimate loop currently runs as fast as possible, and outputs
--   a new command on every iteration. Should probably do something
--   slightly better than that, cf. the old Cyebot code.
-- * Should also somehow catch and handle IO exceptions in the sampler
--   and renderer.
-- * What about termination? Currently runs for ever. Key comand?

pioneerCamera :: VideoSource
pioneerCamera = VSStereoVidere "This configuration string is not used."


reactimatePioneer :: ST PioneerInput PioneerOutput -> IO ()
reactimatePioneer = reactimatePioneerVS pioneerCamera


reactimatePioneerVS :: VideoSource -> ST PioneerInput PioneerOutput -> IO ()
reactimatePioneerVS vs st = do
    -- Start pioneer client/server and initialize the Pioneer
    startServers
    pioneerResetOdometry

    -- Initialize FVision input and output.
    video <- openVideoSource vs
    guiConsole <- openConsoleWindow video (0,0)

    -- Initialize sampler and renderer.
    pi0         <- getPioneerInput video guiConsole
    sysSampler  <- mkSysSampler video guiConsole
    sysRenderer <- mkSysRenderer guiConsole

    -- Main loop; will currently not terminate.
    reactimate st pi0 () idFilter sysSampler sysRenderer

    -- Close down.
    pioneerStop


-- I don't find the filter very useful. The necessary stateful processing
-- can be done in the IO monad anyway. Sometimes it HAS to be done there.
-- Get rid of the filter?
idFilter :: PrimFilter () (DTime, PioneerInput) PioneerInput
idFilter _ (dt, pi) = ((), dt, Just pi, False)


mkSysSampler :: Video -> XVWindow -> IO (SysSampler (DTime, PioneerInput))
mkSysSampler video guiConsole = do
    clkRes <- fmap fromIntegral (getSysVar ClockTick)
    t0     <- getElapsedTime
    tpRef  <- newIORef t0
    let sysSampler _ _ =
	    do
		tp <- readIORef tpRef
		t  <- getElapsedTime `repeatUntil` (> tp)
		let dt = fromIntegral (t - tp) / clkRes
		writeIORef tpRef t
		pi <- getPioneerInput video guiConsole
		return (dt, pi)
    return sysSampler


-- Accurate enough?
getElapsedTime :: IO ClockTick
getElapsedTime = fmap elapsedTime getProcessTimes


-- Gets the input from the various sensors and the (GUI) console.
-- As a side effect, it will also forward the read image to the GUI console.
-- This might be too simplistic, depending on the semantics of the incoming
-- Maybes. If they have a "continuous" character, then this is good enough. 
-- Otherwise, we need more events (like mouse entering/leaving window,
-- button/key up AND down) and stateful processing for generating continuous
-- input signals.
getPioneerInput :: Video -> XVWindow -> IO PioneerInput
getPioneerInput video guiConsole =
    do
        odometry <- pioneerGetPos
        let odometry = (0.0, 0.0, 0.0)
        image    <- getNextImage video
        putImage guiConsole image
        input <- getConsoleWindowInput guiConsole
	return (mkPI odometry image input)
    where
	mkPI (xpos, ypos, theta)
	     image
	     (guiImage, mMousePos, mMouseLBP, mMouseRBP, mKey) =
	    PI {
		piBattStat = BSHigh,		-- For now.
		piIsStuck  = False,		-- For now.
		piPosition = (xpos, ypos),
		piHeading  = theta,
		piImage    = image,
		piKey      = mKey,
		piGUIImage = guiImage,
		piMousePos = mMousePos,
		piMouseLB  = mMouseLBP,
		piMouseRB  = mMouseRBP
	    }


-- A bit too simplistic. Should really avoid sending repetitive commands.
mkSysRenderer :: XVWindow -> IO (SysRenderer PioneerOutput)
mkSysRenderer guiConsole = do
    let sysRenderer' (PO {poDD = ddm, poTCO = msgs, poFVGO = sg}) = do
	    case ddm of
		DDMBrake -> pioneerStop
		DDMDiff {ddmdLWV = lwv, ddmdRWV = rwv} ->
		    pioneerSetVelDiff lwv rwv
		DDMTR {ddmtrTV = v, ddmtrRV = rv} ->
		    pioneerSetVelTR v rv
	    forAll msgs putStrLn
	    sgRender guiConsole sg
	    flushToConsoleWindow guiConsole
	    return ()
    -- The change mechanism is not used in the stream-based AFRP.
    return (\change po -> if change then sysRenderer' po else return ())
