{- $Id: PioneerIO.hs,v 1.1.1.1 2002/06/26 21:40:30 henrik Exp $
******************************************************************************
*                          A F R O B / P I O N E E R                         *
*                                                                            *
*	Module:		PioneerIO					     *
*	Purpose:	Pioneer I/O types and instances.		     *
*	Author:		Henrik Nilsson					     *
*									     *
******************************************************************************
-}

-- Pioneer I/O types for Pioneers equipped with FVision-based (and thus
-- XVision-based) vision and GUI.
--
-- Not quite sure any more it is such a great idea to make individual robot
-- types, even specific configurations of such robots, part of the AFRob
-- framework.
-- The alternative would be to make AFrob quite self contained (including
-- what's now in Common, and possibly a simulator), and moving all this stuff
-- to the application level since it is quite unlikely that this stuff is
-- suficiently general to merrit being put into a library.

module PioneerIO (
    PioneerInput(..),
    PioneerOutput(..),
    DiffDriveMode(..)
) where

import AFRP (Event, event, ComposableOutput(..), coUpdater)
import AFrobPhysicalDimensions
import AFrobRobotIO
import AFVision
import PioneerBridge (pioneerDiameter, pioneerMaxAcceleration, pioneerMaxSpeed)


------------------------------------------------------------------------------
-- Input type and instances
------------------------------------------------------------------------------

data PioneerInput = PI {
    -- Fields for HasRobotStatus
    piBattStat :: BatteryStatus,
    piIsStuck  :: Bool,

    -- Fields for HasOdometry
    piPosition :: Position2,
    piHeading  :: Heading,

    -- Fields for HasFVisionInput
    piImage    :: ImageRGB,

    -- Fields for HasTextualConsoleInput
    piKey      :: Maybe Char,

    -- Fields for HasFVisionGUIInput
    piGUIImage :: ImageRGB,
    piMousePos :: Maybe Point,
    piMouseLB  :: Maybe Point,
    piMouseRB  :: Maybe Point
}


instance HasRobotProperties PioneerInput where
    rpId       = const 0			-- Fake it for now.
    rpDiameter = const pioneerDiameter
    rpAccMax   = const pioneerMaxAcceleration
    rpWSMax    = const pioneerMaxSpeed


instance HasRobotStatus PioneerInput where
    rsBattStat = piBattStat
    rsIsStuck  = piIsStuck


instance HasOdometry PioneerInput where
    odometryPosition = piPosition
    odometryHeading  = piHeading


instance HasTextualConsoleInput PioneerInput where
    tciKey = piKey


instance HasFVisionInput PioneerInput where
    fviImage = piImage


instance HasFVisionGUIInput PioneerInput where
    fvgiGUIImage = piGUIImage
    fvgiMousePos = piMousePos
    fvgiMouseLB  = piMouseLB
    fvgiMouseRB  = piMouseRB


------------------------------------------------------------------------------
-- Output type with instances and subordinate types
------------------------------------------------------------------------------

data DiffDriveMode =
      DDMBrake				-- Brake both wheels.
    | DDMDiff {
	  ddmdLWV :: Velocity,		-- Left Wheel Velocity.
	  ddmdRWV :: Velocity		-- Right Wheel Velocity.
      }
    | DDMTR {
	  ddmtrTV :: Velocity,		-- Translational Velocity.
	  ddmtrRV :: RotVel		-- Rotational Velocity.
      }


data PioneerOutput = PO {
    poDD   :: DiffDriveMode,		-- Differential drive output.
    poTCO  :: [String],			-- Text console output.
    poFVGO :: SimpleGraphic		-- FVision GUI output.
}


instance ComposableOutput PioneerOutput where
    coDefault = PO {poDD = DDMBrake, poTCO = [], poFVGO = SGEmpty}


instance HasDiffDrive PioneerOutput where
    ddSetBrake = coUpdater (\o -> o {poDD = DDMBrake})

    ddSetVelDiff lwv rwv =
        coUpdater (\o -> o {poDD = DDMDiff {ddmdLWV = lwv, ddmdRWV = rwv}})

    ddSetVelTR v rv =
        coUpdater (\o -> o {poDD = DDMTR {ddmtrTV = v, ddmtrRV = rv}})


instance HasTextConsoleOutput PioneerOutput where
    tcoAddMessage em =
        coUpdater (\o -> event o (\m -> o {poTCO = m : poTCO o}) em)


instance HasFVisionGUIOutput PioneerOutput where
    fvgoOverlaySimpleGraphic sg =
	coUpdater (\o -> o {poFVGO = sg `SGOver` poFVGO o})
